# Copyright (c) 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import datetime

from sqlalchemy import asc
from sqlalchemy import func
from sqlalchemy import or_
from sqlalchemy.orm import aliased

from networking_odl.common import constants as odl_const
from networking_odl.db import models

from neutron.db import api as db_api

from oslo_db import api as oslo_db_api
from oslo_log import log as logging


LOG = logging.getLogger(__name__)


def get_pending_or_processing_ops(session, object_uuid, operation=None):
    q = session.query(models.OpenDaylightJournal).filter(
        or_(models.OpenDaylightJournal.state == odl_const.PENDING,
            models.OpenDaylightJournal.state == odl_const.PROCESSING),
        models.OpenDaylightJournal.object_uuid == object_uuid)

    if operation:
        if isinstance(operation, (list, tuple)):
            q = q.filter(models.OpenDaylightJournal.operation.in_(operation))
        else:
            q = q.filter(models.OpenDaylightJournal.operation == operation)

    return q.all()


def get_pending_delete_ops_with_parent(session, object_type, parent_id):
    rows = session.query(models.OpenDaylightJournal).filter(
        or_(models.OpenDaylightJournal.state == odl_const.PENDING,
            models.OpenDaylightJournal.state == odl_const.PROCESSING),
        models.OpenDaylightJournal.object_type == object_type,
        models.OpenDaylightJournal.operation == odl_const.ODL_DELETE
    ).all()

    return (row for row in rows if parent_id in row.data)


def get_all_db_rows(session):
    return session.query(models.OpenDaylightJournal).all()


def get_all_db_rows_by_state(session, state):
    return session.query(models.OpenDaylightJournal).filter_by(
        state=state).all()


# Retry deadlock exception for Galera DB.
# If two (or more) different threads call this method at the same time, they
# might both succeed in changing the same row to pending, but at least one
# of them will get a deadlock from Galera and will have to retry the operation.
@db_api.retry_db_errors
def get_oldest_pending_db_row_with_lock(session):
    with session.begin():
        journal_dep = aliased(models.OpenDaylightJournal)
        dep_query = session.query(journal_dep).filter(
            models.OpenDaylightJournal.seqnum == journal_dep.seqnum
        ).outerjoin(
            journal_dep.depending_on, aliased=True).filter(
            or_(models.OpenDaylightJournal.state == odl_const.PENDING,
                models.OpenDaylightJournal.state == odl_const.PROCESSING))
        row = session.query(models.OpenDaylightJournal).filter(
            models.OpenDaylightJournal.state == odl_const.PENDING,
            ~ dep_query.exists()
        ).order_by(
            asc(models.OpenDaylightJournal.last_retried)).first()
        if row:
            update_db_row_state(session, row, odl_const.PROCESSING)

    return row


def delete_dependency(session, entry):
    """Delete dependency upon the given ID"""
    conn = session.connection()
    stmt = models.journal_dependencies.delete(
        models.journal_dependencies.c.depends_on == entry.seqnum)
    conn.execute(stmt)
    session.expire_all()


@oslo_db_api.wrap_db_retry(max_retries=db_api.MAX_RETRIES)
def update_db_row_state(session, row, state, flush=True):
    row.state = state
    session.merge(row)
    if flush:
        session.flush()


def update_pending_db_row_retry(session, row, retry_count):
    if row.retry_count >= retry_count:
        update_db_row_state(session, row, odl_const.FAILED)
    else:
        row.retry_count += 1
        update_db_row_state(session, row, odl_const.PENDING)


@oslo_db_api.wrap_db_retry(max_retries=db_api.MAX_RETRIES)
def delete_row(session, row=None, row_id=None, flush=True):
    if row_id:
        row = session.query(models.OpenDaylightJournal).filter_by(
            seqnum=row_id).one()
    if row:
        session.delete(row)
        if flush:
            session.flush()


@oslo_db_api.wrap_db_retry(max_retries=db_api.MAX_RETRIES)
def create_pending_row(session, object_type, object_uuid,
                       operation, data, depending_on=None):
    if depending_on is None:
        depending_on = []
    row = models.OpenDaylightJournal(object_type=object_type,
                                     object_uuid=object_uuid,
                                     operation=operation, data=data,
                                     state=odl_const.PENDING,
                                     depending_on=depending_on)
    session.add(row)
    # Keep session flush for unit tests. NOOP for L2/L3 events since calls are
    # made inside database session transaction with subtransactions=True.
    session.flush()


@db_api.retry_db_errors
def delete_pending_rows(session, operations_to_delete):
    with session.begin():
        session.query(models.OpenDaylightJournal).filter(
            models.OpenDaylightJournal.operation.in_(operations_to_delete),
            models.OpenDaylightJournal.state == odl_const.PENDING).delete(
            synchronize_session=False)
        session.expire_all()


@db_api.retry_db_errors
def _update_periodic_task_state(session, expected_state, state, task):
    with session.begin():
        row = session.query(models.OpenDaylightPeriodicTask).filter_by(
            state=expected_state,
            task=task).with_for_update().one_or_none()

        if row is None:
            return False

        row.state = state
        return True


def was_periodic_task_executed_recently(session, task, interval):
    now = session.execute(func.now()).scalar()
    delta = datetime.timedelta(seconds=interval)
    row = session.query(models.OpenDaylightPeriodicTask).filter(
        models.OpenDaylightPeriodicTask.task == task,
        (now - delta >= (models.OpenDaylightPeriodicTask.lock_updated))
    ).one_or_none()

    return bool(row is None)


def lock_periodic_task(session, task):
    return _update_periodic_task_state(session, odl_const.PENDING,
                                       odl_const.PROCESSING, task)


def unlock_periodic_task(session, task):
    return _update_periodic_task_state(session, odl_const.PROCESSING,
                                       odl_const.PENDING, task)


def update_periodic_task(session, task, operation=None):
    """Update the current periodic task details.

    The function assumes the lock is held, so it mustn't be run outside of a
    locked context.
    """
    op_text = None
    if operation:
        op_text = operation.__name__

    with session.begin():
        row = session.query(models.OpenDaylightPeriodicTask).filter_by(
            task=task).one()
        row.processing_operation = op_text


def delete_rows_by_state_and_time(session, state, time_delta):
    # NOTE(mpeterson): The reason behind deleting one-by-one is that InnoDB
    # ignores the WHERE clause to issue a LOCK when executing a DELETE. By
    # executing each operation indepently, we minimize exposures to DEADLOCKS.
    with session.begin():
        now = session.execute(func.now()).scalar()
        rows = session.query(models.OpenDaylightJournal).filter(
            models.OpenDaylightJournal.state == state,
            models.OpenDaylightJournal.last_retried < now - time_delta).all()
        for row in rows:
            delete_row(session, row, flush=False)
        session.expire_all()


def reset_processing_rows(session, max_timedelta):
    # NOTE(mpeterson): The reason behind updating one-by-one is that InnoDB
    # ignores the WHERE clause to issue a LOCK when executing an UPDATE. By
    # executing each operation indepently, we minimize exposures to DEADLOCKS.
    with session.begin():
        now = session.execute(func.now()).scalar()
        max_timedelta = datetime.timedelta(seconds=max_timedelta)
        rows = session.query(models.OpenDaylightJournal).filter(
            models.OpenDaylightJournal.last_retried < now - max_timedelta,
            models.OpenDaylightJournal.state == odl_const.PROCESSING).all()
        for row in rows:
            update_db_row_state(session, row, odl_const.PENDING, flush=False)

    return len(rows)



# Copyright (c) 2017 Ericsson India Global Service Pvt Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from neutron.tests import base as base_test

from networking_odl.common.client import OpenDaylightRestClient as client
from networking_odl.common import constants as odl_const
from networking_odl.tests import base as odl_base
from networking_odl.trunk import trunk_driver_v1 as trunk_driver

from neutron.services.trunk import callbacks
from neutron.services.trunk import constants as trunk_consts

from neutron_lib.callbacks import events
from oslo_config import cfg


FAKE_TRUNK = {
    'status': 'ACTIVE',
    'sub_ports': [{'segmentation_type': 'vlan',
                   'port_id': 'fake_port_id',
                   'segmentation_id': 101},
                  {'segmentation_type': 'vlan',
                   'port_id': 'fake_port_id',
                   'segmentation_id': 102}],
    'name': 'trunk0',
    'admin_state_up': 'true',
    'tenant_id': 'fake_tenant_id',
    'updated_at': '2016-11-16T10:17:44Z',
    'revision_number': 2,
    'project_id': 'fake_project_id',
    'port_id': 'fake_port_id',
    'id': 'fake_id',
    'description': 'fake trunk port'}


class TestTrunkHandler(base_test.BaseTestCase):
    def setUp(self):
        self.useFixture(odl_base.OpenDaylightRestClientFixture())
        super(TestTrunkHandler, self).setUp()
        self.handler = (trunk_driver.
                        OpenDaylightTrunkHandlerV1())

    def _fake_trunk_event_payload(self):
        payload = callbacks.TrunkPayload(
            mock.Mock(), 'fake_id',
            mock.Mock(return_value=FAKE_TRUNK),
            mock.Mock(return_value=FAKE_TRUNK),
            mock.Mock(return_value=FAKE_TRUNK['sub_ports']))
        payload.current_trunk.status = trunk_consts.DOWN_STATUS
        payload.current_trunk.to_dict = mock.Mock(return_value=FAKE_TRUNK)
        payload.original_trunk.status = trunk_consts.DOWN_STATUS
        payload.original_trunk.to_dict = mock.Mock(return_value=FAKE_TRUNK)
        return payload

    @mock.patch.object(client, 'sendjson')
    def test_create_trunk(self, mocked_sendjson):
        fake_payload = self._fake_trunk_event_payload()
        expected = {odl_const.ODL_TRUNK: fake_payload.current_trunk.to_dict()}
        self.handler.trunk_event(mock.ANY, events.AFTER_CREATE,
                                 mock.ANY, fake_payload)
        mocked_sendjson.assert_called_once_with('post', odl_const.ODL_TRUNKS,
                                                expected)

    @mock.patch.object(client, 'sendjson')
    def test_update_trunk(self, mocked_sendjson):
        fake_payload = self._fake_trunk_event_payload()
        expected = {odl_const.ODL_TRUNK: fake_payload.current_trunk.to_dict()}
        self.handler.trunk_event(mock.ANY, events.AFTER_UPDATE,
                                 mock.ANY, fake_payload)
        url = odl_const.ODL_TRUNKS + '/' + fake_payload.trunk_id
        mocked_sendjson.assert_called_once_with('put', url, expected)

    @mock.patch.object(client, 'sendjson')
    def test_subport(self, mocked_sendjson):
        fake_payload = self._fake_trunk_event_payload()
        expected = {odl_const.ODL_TRUNK: fake_payload.current_trunk.to_dict()}
        self.handler.subport_event(mock.ANY, mock.ANY, mock.ANY, fake_payload)
        url = odl_const.ODL_TRUNKS + '/' + fake_payload.trunk_id
        mocked_sendjson.assert_called_once_with('put', url, expected)

    @mock.patch.object(client, 'try_delete')
    def test_delete_trunk(self, mocked_try_delete):
        fake_payload = self._fake_trunk_event_payload()
        self.handler.trunk_event(mock.ANY, events.AFTER_DELETE,
                                 mock.ANY, fake_payload)
        url = odl_const.ODL_TRUNKS + '/' + fake_payload.trunk_id
        mocked_try_delete.assert_called_once_with(url)


class TestTrunkDriver(base_test.BaseTestCase):
    def setUp(self):
        self.useFixture(odl_base.OpenDaylightRestClientFixture())
        super(TestTrunkDriver, self).setUp()

    def test_is_loaded(self):
        driver = trunk_driver.OpenDaylightTrunkDriverV1.create()
        cfg.CONF.set_override('mechanism_drivers',
                              ["logger", odl_const.ODL_ML2_MECH_DRIVER_V1],
                              group='ml2')
        self.assertTrue(driver.is_loaded)

        cfg.CONF.set_override('mechanism_drivers',
                              ['logger'],
                              group='ml2')
        self.assertFalse(driver.is_loaded)

        cfg.CONF.set_override('core_plugin', 'some_plugin')
        self.assertFalse(driver.is_loaded)



# Copyright (c) 2016 Brocade Communication Systems
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from networking_sfc.services.sfc.drivers import base as sfc_driver

from networking_odl.common import client as odl_client
from networking_odl.common import constants as odl_const

LOG = logging.getLogger(__name__)


class OpenDaylightSFCDriverV1(sfc_driver.SfcDriverBase):
    """OpenDaylight SFC Driver for networking-sfc.

    Driver sends REST request for Networking SFC Resources (Port Pair,
    Port Pair Group & Port Chain) to OpenDaylight Neutron Northbound.
    OpenDaylight Neutron Northbound has API's defined for these resources
    based on the Networking SFC APIs.
    """

    def initialize(self):
        LOG.debug("Initializing OpenDaylight Networking SFC driver")
        self.client = odl_client.OpenDaylightRestClient.create_client()

    @log_helpers.log_method_call
    def create_port_pair(self, context):
        self.client.send_request(odl_const.ODL_CREATE,
                                 odl_const.ODL_SFC,
                                 odl_const.ODL_SFC_PORT_PAIR,
                                 context.current)

    @log_helpers.log_method_call
    def update_port_pair(self, context):
        self.client.send_request(odl_const.ODL_UPDATE,
                                 odl_const.ODL_SFC,
                                 odl_const.ODL_SFC_PORT_PAIR,
                                 context.current)

    @log_helpers.log_method_call
    def delete_port_pair(self, context):
        self.client.send_request(odl_const.ODL_DELETE,
                                 odl_const.ODL_SFC,
                                 odl_const.ODL_SFC_PORT_PAIR,
                                 context.current)

    @log_helpers.log_method_call
    def create_port_pair_group(self, context):
        self.client.send_request(odl_const.ODL_CREATE,
                                 odl_const.ODL_SFC,
                                 odl_const.ODL_SFC_PORT_PAIR_GROUP,
                                 context.current)

    @log_helpers.log_method_call
    def update_port_pair_group(self, context):
        self.client.send_request(odl_const.ODL_UPDATE,
                                 odl_const.ODL_SFC,
                                 odl_const.ODL_SFC_PORT_PAIR_GROUP,
                                 context.current)

    @log_helpers.log_method_call
    def delete_port_pair_group(self, context):
        self.client.send_request(odl_const.ODL_DELETE,
                                 odl_const.ODL_SFC,
                                 odl_const.ODL_SFC_PORT_PAIR_GROUP,
                                 context.current)

    @log_helpers.log_method_call
    def create_port_chain(self, context):
        self.client.send_request(odl_const.ODL_CREATE,
                                 odl_const.ODL_SFC,
                                 odl_const.ODL_SFC_PORT_CHAIN,
                                 context.current)

    @log_helpers.log_method_call
    def update_port_chain(self, context):
        self.client.send_request(odl_const.ODL_UPDATE,
                                 odl_const.ODL_SFC,
                                 odl_const.ODL_SFC_PORT_CHAIN,
                                 context.current)

    @log_helpers.log_method_call
    def delete_port_chain(self, context):
        self.client.send_request(odl_const.ODL_DELETE,
                                 odl_const.ODL_SFC,
                                 odl_const.ODL_SFC_PORT_CHAIN,
                                 context.current)



# Copyright (c) 2017 Ericsson India Global Service Pvt Ltd.
# All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils

from neutron.services.trunk import constants as t_consts
from neutron.services.trunk.drivers import base as trunk_base

from networking_odl.common import client as odl_client
from networking_odl.common import constants as odl_const
from networking_odl.trunk import constants as odltrunk_const


LOG = logging.getLogger(__name__)

# NOTE: Status handling
# V1 driver assumes if status=ACTIVE by default and sets it before making
# Create/Update rest calls to ODL.
# In case of failure in rest, it resets it to DEGRADED.


@registry.has_registry_receivers
class OpenDaylightTrunkHandlerV1(object):
    def __init__(self):
        self.client = odl_client.OpenDaylightRestClient.create_client()
        LOG.info('initialized trunk driver for OpendayLight')

    def trunk_create_postcommit(self, trunk):
        trunk.update(status=t_consts.ACTIVE_STATUS)
        trunk_dict = trunk.to_dict()
        try:
            self.client.sendjson('post', odl_const.ODL_TRUNKS,
                                 {odl_const.ODL_TRUNK: trunk_dict})
        except Exception:
            with excutils.save_and_reraise_exception():
                trunk.update(status=t_consts.DEGRADED_STATUS)

    def trunk_delete_postcommit(self, trunk):
        trunk_dict = trunk.to_dict()
        url = odl_const.ODL_TRUNKS + '/' + trunk_dict['id']
        self.client.try_delete(url)

    def trunk_update_postcommit(self, updated):
        updated.update(status=t_consts.ACTIVE_STATUS)
        trunk_dict = updated.to_dict()
        try:
            url = odl_const.ODL_TRUNKS + '/' + trunk_dict['id']
            self.client.sendjson('put', url,
                                 {odl_const.ODL_TRUNK: trunk_dict})
        except Exception:
            with excutils.save_and_reraise_exception():
                updated.update(status=t_consts.DEGRADED_STATUS)

    @registry.receives(t_consts.TRUNK, (events.AFTER_CREATE,
                                        events.AFTER_DELETE,
                                        events.AFTER_UPDATE))
    def trunk_event(self, resource, event, trunk_plugin, payload):
        if event == events.AFTER_CREATE:
            self.trunk_create_postcommit(payload.current_trunk)
        if event == events.AFTER_UPDATE:
            self.trunk_update_postcommit(payload.current_trunk)
        elif event == events.AFTER_DELETE:
            self.trunk_delete_postcommit(payload.original_trunk)

    @registry.receives(t_consts.SUBPORTS, (events.AFTER_CREATE,
                                           events.AFTER_DELETE))
    def subport_event(self, resource, event, trunk_plugin, payload):
        self.trunk_update_postcommit(payload.current_trunk)


class OpenDaylightTrunkDriverV1(trunk_base.DriverBase):
    @property
    def is_loaded(self):
        try:
            return (odl_const.ODL_ML2_MECH_DRIVER_V1 in
                    cfg.CONF.ml2.mechanism_drivers)
        except cfg.NoSuchOptError:
            return False

    @registry.receives(t_consts.TRUNK_PLUGIN, [events.AFTER_INIT])
    def register(self, resource, event, trigger, **kwargs):
        super(OpenDaylightTrunkDriverV1, self).register(
            resource, event, trigger, **kwargs)
        self._handler = OpenDaylightTrunkHandlerV1()

    @classmethod
    def create(cls):
        return cls(odl_const.ODL_ML2_MECH_DRIVER_V1,
                   odltrunk_const.SUPPORTED_INTERFACES,
                   odltrunk_const.SUPPORTED_SEGMENTATION_TYPES,
                   None,
                   can_trunk_bound_port=True)



# Copyright (c) 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from neutron.tests import base

from networking_odl.common import utils


class TestUtils(base.DietTestCase):

    # TODO(manjeets) remove this test once neutronify is
    # consolidated with make_plural
    def test_neutronify(self):
        self.assertEqual('a-b-c', utils.neutronify('a_b_c'))

    def test_neutronify_empty(self):
        self.assertEqual('', utils.neutronify(''))

    def test_make_url_object_in_resource_map(self):
        url_object = utils.make_url_object('policy')
        self.assertEqual('qos/policies', url_object)

    def test_make_url_sfc_object_in_resource_map(self):
        objs = ['flowclassifier', 'portpair', 'portpairgroup', 'portchain']
        for obj in objs:
            url_object = utils.make_url_object(obj)
            self.assertEqual('sfc/%ss' % obj, url_object)

    def test_make_url_object_conversion(self):
        self.assertEqual('networks', utils.make_url_object('network'))
        self.assertEqual('l2-gateways', utils.make_url_object('l2_gateway'))

    def test_get_odl_url(self):
        """test make uri."""
        cfg.CONF.set_override('url',
                              'http://localhost:8080'
                              '/controller/nb/v2/neutron', 'ml2_odl')
        test_path = '/restconf/neutron:neutron/hostconfigs'
        expected = "http://localhost:8080/restconf/neutron:neutron/hostconfigs"
        test_uri = utils.get_odl_url(path=test_path)

        self.assertEqual(expected, test_uri)



# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from oslo_config import cfg
from oslo_serialization import jsonutils
from requests import exceptions

from networking_odl.common.client import OpenDaylightRestClient
from networking_odl.common import odl_features
from networking_odl.tests import base


class TestOdlFeatures(base.DietTestCase):
    """Basic tests for odl_features"""

    feature_json = """{"features": {"feature":
                            [{"service-provider-feature":
                            "neutron-extensions:operational-port-status"}]}}"""

    def setUp(self):
        self.features_fixture = base.OpenDaylightFeaturesFixture()
        self.useFixture(self.features_fixture)
        super(TestOdlFeatures, self).setUp()
        self.features_fixture.mock_odl_features_init.stop()

    @mock.patch.object(OpenDaylightRestClient, 'request')
    def test_fetch_exception(self, mocked_client):
        mocked_client.side_effect = exceptions.ConnectionError()
        self.assertTrue(None == odl_features._fetch_features())

    @mock.patch.object(OpenDaylightRestClient, 'request')
    def test_fetch_404(self, mocked_client):
        mocked_client.return_value = mock.MagicMock(status_code=404)
        self.assertTrue(set() == odl_features._fetch_features())

    @mock.patch.object(OpenDaylightRestClient, 'request')
    def test_fetch_400(self, mocked_client):
        mocked_client.return_value = mock.MagicMock(status_code=400)
        self.assertTrue(set() == odl_features._fetch_features())

    @mock.patch.object(OpenDaylightRestClient, 'request')
    def test_fetch_500(self, mocked_client):
        mocked_client.return_value = mock.MagicMock(status_code=500)
        self.assertTrue(None == odl_features._fetch_features())

    @mock.patch.object(OpenDaylightRestClient, 'request')
    def test_init(self, mocked_client):
        response = mock.MagicMock()
        response.status_code = 200
        response.json = mock.MagicMock(
            return_value=jsonutils.loads(self.feature_json))
        mocked_client.return_value = response

        odl_features.init()
        self.assertTrue(odl_features.has(odl_features.OPERATIONAL_PORT_STATUS))

    def test_init_from_config(self):
        cfg.CONF.set_override('odl_features', 'thing1,thing2', 'ml2_odl')
        odl_features.init()
        self.assertTrue(odl_features.has('thing1'))
        self.assertTrue(odl_features.has('thing2'))



# Copyright (c) 2013-2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from neutron.tests.unit.plugins.ml2 import test_plugin
from neutron_lib import context

from networking_odl.common import constants as const
from networking_odl.ml2 import mech_driver as driver


class TestODLShim(test_plugin.Ml2PluginV2TestCase):

    def setUp(self):
        super(TestODLShim, self).setUp()
        self.context = context.get_admin_context()
        self.plugin = mock.Mock()
        self.driver = driver.OpenDaylightMechanismDriver()
        self.driver.odl_drv = mock.Mock()

    def test_create_network_postcommit(self):
        self.driver.create_network_postcommit(self.context)
        self.driver.odl_drv.synchronize.assert_called_with(const.ODL_CREATE,
                                                           const.ODL_NETWORKS,
                                                           self.context)

    def test_update_network_postcommit(self):
        self.driver.update_network_postcommit(self.context)
        self.driver.odl_drv.synchronize.assert_called_with(const.ODL_UPDATE,
                                                           const.ODL_NETWORKS,
                                                           self.context)

    def test_delete_network_postcommit(self):
        self.driver.delete_network_postcommit(self.context)
        self.driver.odl_drv.synchronize.assert_called_with(const.ODL_DELETE,
                                                           const.ODL_NETWORKS,
                                                           self.context)

    def test_create_subnet_postcommit(self):
        self.driver.create_subnet_postcommit(self.context)
        self.driver.odl_drv.synchronize.assert_called_with(const.ODL_CREATE,
                                                           const.ODL_SUBNETS,
                                                           self.context)

    def test_update_subnet_postcommit(self):
        self.driver.update_subnet_postcommit(self.context)
        self.driver.odl_drv.synchronize.assert_called_with(const.ODL_UPDATE,
                                                           const.ODL_SUBNETS,
                                                           self.context)

    def test_delete_subnet_postcommit(self):
        self.driver.delete_subnet_postcommit(self.context)
        self.driver.odl_drv.synchronize.assert_called_with(const.ODL_DELETE,
                                                           const.ODL_SUBNETS,
                                                           self.context)

    def test_create_port_postcommit(self):
        self.driver.create_port_postcommit(self.context)
        self.driver.odl_drv.synchronize.assert_called_with(const.ODL_CREATE,
                                                           const.ODL_PORTS,
                                                           self.context)

    def test_update_port_postcommit(self):
        self.driver.update_port_postcommit(self.context)
        self.driver.odl_drv.synchronize.assert_called_with(const.ODL_UPDATE,
                                                           const.ODL_PORTS,
                                                           self.context)

    def test_delete_port_postcommit(self):
        self.driver.delete_port_postcommit(self.context)
        self.driver.odl_drv.synchronize.assert_called_with(const.ODL_DELETE,
                                                           const.ODL_PORTS,
                                                           self.context)

    def test_bind_port_delegation(self):
        # given front-end with attached back-end
        front_end = self.driver
        front_end.odl_drv = back_end = mock.MagicMock(
            spec=driver.OpenDaylightDriver)
        # given PortContext to be forwarded to back-end without using
        context = object()

        # when binding port
        front_end.bind_port(context)

        # then port is bound by back-end
        back_end.bind_port.assert_called_once_with(context)



# Copyright 2016 Intel Corporation.
# Copyright 2016 Isaku Yamahata <isaku.yamahata at intel com>
#                               <isaku.yamahata at gmail com>
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.tests.unit.testlib_api import SqlTestCaseLight
from neutron_lib import context

from networking_odl.common import constants
from networking_odl.db import models


class ODLBaseDbTestCase(SqlTestCaseLight):
    def setUp(self):
        super(ODLBaseDbTestCase, self).setUp()
        self.db_context = context.get_admin_context()
        self.db_session = self.db_context.session
        self.addCleanup(self._db_cleanup)

    def _db_cleanup(self):
        self.db_session.query(models.OpenDaylightJournal).delete()
        self.db_session.query(models.OpenDaylightPeriodicTask).delete()
        row0 = models.OpenDaylightPeriodicTask(
            task='maintenance', state=constants.PENDING)
        row1 = models.OpenDaylightPeriodicTask(
            task='hostconfig', state=constants.PENDING)
        self.db_session.merge(row0)
        self.db_session.merge(row1)
        self.db_session.flush()



# Copyright (c) 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from neutron.plugins.ml2.drivers import type_vxlan  # noqa
from neutron import privileged  # noqa
from neutron.tests import base as tests_base
from neutron.tests.common import helpers
from neutron.tests.unit.plugins.ml2 import test_plugin
from oslo_config import cfg

from networking_odl.common import config as odl_config
import networking_odl.db.models  # noqa


class TestODLFullStackBase(test_plugin.Ml2PluginV2TestCase):

    _mechanism_drivers = ['logger', 'opendaylight_v2']
    _extension_drivers = ['port_security']
    l3_plugin = 'networking_odl.l3.l3_odl_v2.OpenDaylightL3RouterPlugin'

    # this is stolen from neutron.tests.fullstack.base
    #
    # This is the directory from which infra fetches log files
    # for fullstack tests
    DEFAULT_LOG_DIR = os.path.join(helpers.get_test_log_path(),
                                   'dsvm-fullstack-logs')

    def setUp(self):
        # NOTE(yamahata):
        # When tox is using virtualenv and oslo.privsep is also installed
        # in system, sudo chooses system one.
        # However for privsep-helper to find python module in virtualenv
        # we need to use the on under .tox. So specify the full path
        # and preserve environmental variables.
        privsep_helper = "privsep-helper"
        if "VIRTUAL_ENV" in os.environ:
            privsep_helper = os.path.join(
                os.environ["VIRTUAL_ENV"], "bin", privsep_helper)
        cfg.CONF.set_override(
            "helper_command", "sudo --preserve-env " + privsep_helper,
            group="privsep")

        cfg.CONF.set_override('extension_drivers',
                              self._extension_drivers, group='ml2')
        cfg.CONF.set_override('tenant_network_types',
                              ['vxlan'], group='ml2')
        cfg.CONF.set_override('vni_ranges',
                              ['1:1000'], group='ml2_type_vxlan')

        odl_url = 'http://127.0.0.1:8087/controller/nb/v2/neutron'
        odl_config.cfg.CONF.set_override('url',
                                         odl_url,
                                         group='ml2_odl')
        odl_config.cfg.CONF.set_override('username',
                                         'admin',
                                         group='ml2_odl')
        odl_config.cfg.CONF.set_override('password',
                                         'admin',
                                         group='ml2_odl')
        odl_config.cfg.CONF.set_override('port_binding_controller',
                                         'legacy-port-binding',
                                         group='ml2_odl')
        odl_config.cfg.CONF.set_override('odl_features',
                                         ['no-feature'],
                                         group='ml2_odl')
        super(TestODLFullStackBase, self).setUp()
        tests_base.setup_test_logging(
            cfg.CONF, self.DEFAULT_LOG_DIR, '%s.txt' % self.get_name())

    def get_name(self):
        class_name, test_name = self.id().split(".")[-2:]
        return "%s.%s" % (class_name, test_name)



# Copyright (c) 2013-2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import copy

import six

import netaddr
from neutron_lib.api.definitions import provider_net as providernet
from neutron_lib.callbacks import resources
from neutron_lib import constants as p_const
from neutron_lib import context as neutron_context
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins.ml2 import api
from neutron_lib.utils import runtime
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
import requests

from neutron.extensions import allowedaddresspairs as addr_pair
from neutron.extensions import multiprovidernet as mpnet
from neutron.extensions import securitygroup as sg
from neutron.plugins.ml2 import driver_context

from networking_odl.common import callback as odl_call
from networking_odl.common import client as odl_client
from networking_odl.common import constants as odl_const
from networking_odl.common import filters
from networking_odl.common import odl_features
from networking_odl.common import utils as odl_utils
from networking_odl.ml2 import port_binding
from networking_odl.ml2 import port_status_update
from networking_odl.trunk import trunk_driver_v1 as trunk_driver


cfg.CONF.import_group('ml2_odl', 'networking_odl.common.config')
LOG = logging.getLogger(__name__)

not_found_exception_map = {odl_const.ODL_NETWORKS: n_exc.NetworkNotFound,
                           odl_const.ODL_SUBNETS: n_exc.SubnetNotFound,
                           odl_const.ODL_PORTS: n_exc.PortNotFound,
                           odl_const.ODL_SGS: sg.SecurityGroupNotFound,
                           odl_const.ODL_SG_RULES:
                               sg.SecurityGroupRuleNotFound}


@six.add_metaclass(abc.ABCMeta)
class ResourceFilterBase(object):
    @staticmethod
    @abc.abstractmethod
    def filter_create_attributes(resource, context):
        pass

    @staticmethod
    @abc.abstractmethod
    def filter_update_attributes(resource, context):
        pass

    @staticmethod
    @abc.abstractmethod
    def filter_create_attributes_with_plugin(resource, plugin, dbcontext):
        pass

    @staticmethod
    def _filter_unmapped_null(resource_dict, unmapped_keys):
        # NOTE(yamahata): bug work around
        # https://bugs.eclipse.org/bugs/show_bug.cgi?id=475475
        #   Null-value for an unmapped element causes next mapped
        #   collection to contain a null value
        #   JSON: { "unmappedField": null, "mappedCollection": [ "a" ] }
        #
        #   Java Object:
        #   class Root {
        #     Collection<String> mappedCollection = new ArrayList<String>;
        #   }
        #
        #   Result:
        #   Field B contains one element; null
        #
        # TODO(yamahata): update along side with neutron and ODL
        #   add when neutron adds more extensions
        #   delete when ODL neutron northbound supports it
        # TODO(yamahata): do same thing for other resources
        keys_to_del = [key for key in unmapped_keys
                       if resource_dict.get(key) is None]
        if keys_to_del:
            odl_utils.try_del(resource_dict, keys_to_del)


class NetworkFilter(ResourceFilterBase):
    _UNMAPPED_KEYS = ['qos_policy_id']

    @classmethod
    def filter_create_attributes(cls, network, context):
        """Filter out network attributes not required for a create."""
        odl_utils.try_del(network, ['status', 'subnets'])
        cls._filter_unmapped_null(network, cls._UNMAPPED_KEYS)

    @classmethod
    def filter_update_attributes(cls, network, context):
        """Filter out network attributes for an update operation."""
        odl_utils.try_del(network, ['id', 'status', 'subnets', 'tenant_id'])
        cls._filter_unmapped_null(network, cls._UNMAPPED_KEYS)

    @classmethod
    def filter_create_attributes_with_plugin(cls, network, plugin, dbcontext):
        context = driver_context.NetworkContext(plugin, dbcontext, network)
        cls.filter_create_attributes(network, context)


class SubnetFilter(ResourceFilterBase):
    _UNMAPPED_KEYS = ['segment_id', 'subnetpool_id']

    @classmethod
    def filter_create_attributes(cls, subnet, context):
        """Filter out subnet attributes not required for a create."""
        cls._filter_unmapped_null(subnet, cls._UNMAPPED_KEYS)

    @classmethod
    def filter_update_attributes(cls, subnet, context):
        """Filter out subnet attributes for an update operation."""
        odl_utils.try_del(subnet, ['id', 'network_id', 'ip_version', 'cidr',
                          'allocation_pools', 'tenant_id'])
        cls._filter_unmapped_null(subnet, cls._UNMAPPED_KEYS)

    @classmethod
    def filter_create_attributes_with_plugin(cls, subnet, plugin, dbcontext):
        network = plugin.get_network(dbcontext, subnet['network_id'])
        context = driver_context.SubnetContext(plugin, dbcontext, subnet,
                                               network)
        cls.filter_create_attributes(subnet, context)
        cls._filter_unmapped_null(subnet, cls._UNMAPPED_KEYS)


class PortFilter(ResourceFilterBase):
    _UNMAPPED_KEYS = ['binding:profile', 'dns_name',
                      'port_security_enabled', 'qos_policy_id']

    @staticmethod
    def _add_security_groups(port, context):
        """Populate the 'security_groups' field with entire records."""
        dbcontext = context._plugin_context
        groups = [context._plugin.get_security_group(dbcontext, sg)
                  for sg in port['security_groups']]
        port['security_groups'] = groups

    @classmethod
    def _fixup_allowed_ipaddress_pairs(cls, allowed_address_pairs):
        """unify (ip address or network address) into network address"""
        for address_pair in allowed_address_pairs:
            ip_address = address_pair['ip_address']
            network_address = str(netaddr.IPNetwork(ip_address))
            address_pair['ip_address'] = network_address

    @classmethod
    def filter_create_attributes(cls, port, context):
        """Filter out port attributes not required for a create."""
        cls._add_security_groups(port, context)
        cls._fixup_allowed_ipaddress_pairs(port[addr_pair.ADDRESS_PAIRS])
        cls._filter_unmapped_null(port, cls._UNMAPPED_KEYS)
        odl_utils.try_del(port, ['status'])

        # NOTE(yamahata): work around for port creation for router
        # tenant_id=''(empty string) is passed when port is created
        # by l3 plugin internally for router.
        # On the other hand, ODL doesn't accept empty string for tenant_id.
        # In that case, deduce tenant_id from network_id for now.
        # Right fix: modify Neutron so that don't allow empty string
        # for tenant_id even for port for internal use.
        # TODO(yamahata): eliminate this work around when neutron side
        # is fixed
        # assert port['tenant_id'] != ''
        if port['tenant_id'] == '':
            LOG.debug('empty string was passed for tenant_id: %s(port)', port)
            port['tenant_id'] = context._network_context._network['tenant_id']

    @classmethod
    def filter_update_attributes(cls, port, context):
        """Filter out port attributes for an update operation."""
        cls._add_security_groups(port, context)
        cls._fixup_allowed_ipaddress_pairs(port[addr_pair.ADDRESS_PAIRS])
        cls._filter_unmapped_null(port, cls._UNMAPPED_KEYS)
        odl_utils.try_del(port, ['network_id', 'id', 'status', 'tenant_id'])

    @classmethod
    def filter_create_attributes_with_plugin(cls, port, plugin, dbcontext):
        network = plugin.get_network(dbcontext, port['network_id'])
        port_db = plugin._get_port(dbcontext, port['id'])
        context = driver_context.PortContext(
            plugin, dbcontext, port, network, port_db.port_binding, None)
        cls.filter_create_attributes(port, context)


class SecurityGroupFilter(ResourceFilterBase):
    @staticmethod
    def filter_create_attributes(sg, context):
        """Filter out security-group attributes not required for a create."""
        pass

    @staticmethod
    def filter_update_attributes(sg, context):
        """Filter out security-group attributes for an update operation."""
        pass

    @staticmethod
    def filter_create_attributes_with_plugin(sg, plugin, dbcontext):
        pass


class SecurityGroupRuleFilter(ResourceFilterBase):
    @staticmethod
    def filter_create_attributes(sg_rule, context):
        """Filter out sg-rule attributes not required for a create."""
        filters.filter_security_group_rule(sg_rule)

    @staticmethod
    def filter_update_attributes(sg_rule, context):
        """Filter out sg-rule attributes for an update operation."""
        filters.filter_security_group_rule(sg_rule)

    @staticmethod
    def filter_create_attributes_with_plugin(sg_rule, plugin, dbcontext):
        filters.filter_security_group_rule(sg_rule)


class OpenDaylightDriver(object):

    """OpenDaylight Python Driver for Neutron.

    This code is the backend implementation for the OpenDaylight ML2
    MechanismDriver for OpenStack Neutron.
    """
    FILTER_MAP = {
        odl_const.ODL_NETWORKS: NetworkFilter,
        odl_const.ODL_SUBNETS: SubnetFilter,
        odl_const.ODL_PORTS: PortFilter,
        odl_const.ODL_SGS: SecurityGroupFilter,
        odl_const.ODL_SG_RULES: SecurityGroupRuleFilter,
    }
    out_of_sync = True

    def __init__(self):
        LOG.debug("Initializing OpenDaylight ML2 driver")
        self.client = odl_client.OpenDaylightRestClient.create_client()
        self.sec_handler = odl_call.OdlSecurityGroupsHandler(
            None, self.sync_from_callback)
        self.port_binding_controller = port_binding.PortBindingManager.create()
        # TODO(rzang): Each port binding controller should have any necessary
        # parameter passed in from configuration files.
        # BTW, CAP_PORT_FILTER seems being obsoleted.
        # Leave the code commmeted out for now for future reference.
        #
        # self.vif_details = {portbindings.CAP_PORT_FILTER: True}
        # self._network_topology = network_topology.NetworkTopologyManager(
        #     vif_details=self.vif_details)

    def synchronize(self, operation, object_type, context):
        """Synchronize ODL with Neutron following a configuration change."""
        if self.out_of_sync:
            self.sync_full(context._plugin)
            if operation in [odl_const.ODL_UPDATE, odl_const.ODL_DELETE]:
                # NOTE(yamahata): work around that sync_full doesn't know
                # how to handle UPDATE and DELETE at the moment.
                # TODO(yamahata): implement TODOs in sync_full and remove this
                # work around
                self.sync_single_resource(operation, object_type, context)
        else:
            self.sync_single_resource(operation, object_type, context)

    def sync_resources(self, plugin, dbcontext, collection_name):
        """Sync objects from Neutron over to OpenDaylight.

        This will handle syncing networks, subnets, and ports from Neutron to
        OpenDaylight. It also filters out the requisite items which are not
        valid for create API operations.
        """
        filter_cls = self.FILTER_MAP[collection_name]
        to_be_synced = []
        obj_getter = getattr(plugin, 'get_%s' % collection_name)
        if collection_name == odl_const.ODL_SGS:
            resources = obj_getter(dbcontext, default_sg=True)
        else:
            resources = obj_getter(dbcontext)
        for resource in resources:
            try:
                # Convert underscores to dashes in the URL for ODL
                collection_name_url = odl_utils.neutronify(collection_name)
                urlpath = collection_name_url + '/' + resource['id']
                self.client.sendjson('get', urlpath, None)
            except requests.exceptions.HTTPError as e:
                with excutils.save_and_reraise_exception() as ctx:
                    if e.response.status_code == requests.codes.not_found:
                        filter_cls.filter_create_attributes_with_plugin(
                            resource, plugin, dbcontext)
                        to_be_synced.append(resource)
                        ctx.reraise = False
            else:
                # TODO(yamahata): compare result with resource.
                # If they don't match, update it below
                pass

        if to_be_synced:
            key = collection_name[:-1] if len(to_be_synced) == 1 else (
                collection_name)
            # Convert underscores to dashes in the URL for ODL
            collection_name_url = odl_utils.neutronify(collection_name)
            self.client.sendjson('post', collection_name_url,
                                 {key: to_be_synced})

        # https://bugs.launchpad.net/networking-odl/+bug/1371115
        # TODO(yamahata): update resources with unsyned attributes
        # TODO(yamahata): find dangling ODL resouce that was deleted in
        # neutron db

    @runtime.synchronized('odl-sync-full')
    def sync_full(self, plugin):
        """Resync the entire database to ODL.

        Transition to the in-sync state on success.
        Note: we only allow a single thread in here at a time.
        """
        if not self.out_of_sync:
            return
        dbcontext = neutron_context.get_admin_context()
        for collection_name in [odl_const.ODL_SGS,
                                odl_const.ODL_SG_RULES,
                                odl_const.ODL_NETWORKS,
                                odl_const.ODL_SUBNETS,
                                odl_const.ODL_PORTS]:
            self.sync_resources(plugin, dbcontext, collection_name)
        self.out_of_sync = False

    def sync_single_resource(self, operation, object_type, context):
        """Sync over a single resource from Neutron to OpenDaylight.

        Handle syncing a single operation over to OpenDaylight, and correctly
        filter attributes out which are not required for the requisite
        operation (create or update) being handled.
        """
        # Convert underscores to dashes in the URL for ODL
        object_type_url = odl_utils.neutronify(object_type)
        try:
            obj_id = context.current['id']
            if operation == odl_const.ODL_DELETE:
                self.out_of_sync |= not self.client.try_delete(
                    object_type_url + '/' + obj_id)
            else:
                filter_cls = self.FILTER_MAP[object_type]
                if operation == odl_const.ODL_CREATE:
                    urlpath = object_type_url
                    method = 'post'
                    attr_filter = filter_cls.filter_create_attributes
                elif operation == odl_const.ODL_UPDATE:
                    urlpath = object_type_url + '/' + obj_id
                    method = 'put'
                    attr_filter = filter_cls.filter_update_attributes
                resource = copy.deepcopy(context.current)
                attr_filter(resource, context)
                self.client.sendjson(method, urlpath,
                                     {object_type_url[:-1]: resource})
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error("Unable to perform %(operation)s on "
                          "%(object_type)s %(object_id)s",
                          {'operation': operation,
                           'object_type': object_type,
                           'object_id': obj_id})
                self.out_of_sync = True

    def sync_from_callback(self, context, operation, res_type,
                           res_id, resource_dict, **kwrags):
        object_type = odl_utils.neutronify(res_type.plural)
        try:
            if operation == odl_const.ODL_DELETE:
                self.out_of_sync |= not self.client.try_delete(
                    object_type + '/' + res_id)
            else:
                if operation == odl_const.ODL_CREATE:
                    urlpath = object_type
                    method = 'post'
                elif operation == odl_const.ODL_UPDATE:
                    urlpath = object_type + '/' + res_id
                    method = 'put'
                self.client.sendjson(method, urlpath, resource_dict)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error("Unable to perform %(operation)s on "
                          "%(object_type)s %(res_id)s %(resource_dict)s",
                          {'operation': operation,
                           'object_type': object_type,
                           'res_id': res_id,
                           'resource_dict': resource_dict})
                self.out_of_sync = True

        # NOTE(yamahata) when security group is created, default rules
        # are also created.
        if (operation == odl_const.ODL_CREATE and
                res_type.singular == odl_const.ODL_SG):
            for rule in resource_dict[odl_const.ODL_SG][
                    odl_const.ODL_SG_RULES]:
                self.sync_from_callback(
                    context, odl_const.ODL_CREATE,
                    odl_call._RESOURCE_MAPPING[resources.SECURITY_GROUP_RULE],
                    rule['id'], {odl_const.ODL_SG_RULE: rule})

    def bind_port(self, port_context):
        """Set binding for a valid segments

        """
        self.port_binding_controller.bind_port(port_context)

    def check_vlan_transparency(self, context):
        """Check VLAN transparency

        """
        # VLAN and FLAT cases, we don't know if the underlying network
        # supports QinQ or VLAN.
        # For now, netvirt supports only vxlan tunneling.
        VLAN_TRANSPARENT_NETWORK_TYPES = [p_const.TYPE_VXLAN]
        network = context.current
        # see TypeManager._extend_network_dict_provider()
        # single providernet
        if providernet.NETWORK_TYPE in network:
            return (network[providernet.NETWORK_TYPE] in
                    VLAN_TRANSPARENT_NETWORK_TYPES)
        # multi providernet
        segments = network.get(mpnet.SEGMENTS)
        if segments is None:
            return True
        return all(segment[providernet.NETWORK_TYPE]
                   in VLAN_TRANSPARENT_NETWORK_TYPES
                   for segment in segments)


class OpenDaylightMechanismDriver(api.MechanismDriver):

    """Mechanism Driver for OpenDaylight.

    This driver was a port from the NCS MechanismDriver.  The API
    exposed by ODL is slightly different from the API exposed by NCS,
    but the general concepts are the same.
    """

    def initialize(self):
        self.url = cfg.CONF.ml2_odl.url
        self.timeout = cfg.CONF.ml2_odl.timeout
        self.username = cfg.CONF.ml2_odl.username
        self.password = cfg.CONF.ml2_odl.password
        self.odl_drv = OpenDaylightDriver()
        self.trunk_driver = trunk_driver.OpenDaylightTrunkDriverV1.create()
        odl_features.init()

    def get_workers(self):
        workers = [port_status_update.OdlPortStatusUpdate()]
        workers += self.odl_drv.port_binding_controller.get_workers()
        return workers

    # Postcommit hooks are used to trigger synchronization.

    def create_network_postcommit(self, context):
        self.odl_drv.synchronize(odl_const.ODL_CREATE, odl_const.ODL_NETWORKS,
                                 context)

    def update_network_postcommit(self, context):
        self.odl_drv.synchronize(odl_const.ODL_UPDATE, odl_const.ODL_NETWORKS,
                                 context)

    def delete_network_postcommit(self, context):
        self.odl_drv.synchronize(odl_const.ODL_DELETE, odl_const.ODL_NETWORKS,
                                 context)

    def create_subnet_postcommit(self, context):
        self.odl_drv.synchronize(odl_const.ODL_CREATE, odl_const.ODL_SUBNETS,
                                 context)

    def update_subnet_postcommit(self, context):
        self.odl_drv.synchronize(odl_const.ODL_UPDATE, odl_const.ODL_SUBNETS,
                                 context)

    def delete_subnet_postcommit(self, context):
        self.odl_drv.synchronize(odl_const.ODL_DELETE, odl_const.ODL_SUBNETS,
                                 context)

    def create_port_postcommit(self, context):
        self.odl_drv.synchronize(odl_const.ODL_CREATE, odl_const.ODL_PORTS,
                                 context)

    def update_port_postcommit(self, context):
        self.odl_drv.synchronize(odl_const.ODL_UPDATE, odl_const.ODL_PORTS,
                                 context)

    def delete_port_postcommit(self, context):
        self.odl_drv.synchronize(odl_const.ODL_DELETE, odl_const.ODL_PORTS,
                                 context)

    def bind_port(self, context):
        self.odl_drv.bind_port(context)

    def check_vlan_transparency(self, context):
        return self.odl_drv.check_vlan_transparency(context)



#
# Copyright (C) 2013 Red Hat, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from neutron_lbaas.drivers import driver_base
from neutron_lbaas.drivers import driver_mixins

from networking_odl.common import client as odl_client
from networking_odl.common import constants as odl_const

cfg.CONF.import_group('ml2_odl', 'networking_odl.common.config')
LOG = logging.getLogger(__name__)

LBAAS = "lbaas"


class OpenDaylightLbaasDriverV2(driver_base.LoadBalancerBaseDriver):

    @log_helpers.log_method_call
    def __init__(self, plugin):
        LOG.debug("Initializing OpenDaylight LBaaS driver")
        self.plugin = plugin
        self.client = odl_client.OpenDaylightRestClient.create_client()
        self.load_balancer = ODLLoadBalancerManager(self, self.client)
        self.listener = ODLListenerManager(self, self.client)
        self.pool = ODLPoolManager(self, self.client)
        self.member = ODLMemberManager(self, self.client)
        self.health_monitor = ODLHealthMonitorManager(self, self.client)


class OpenDaylightManager(driver_mixins.BaseManagerMixin):

    out_of_sync = True
    url_path = ""
    obj_type = ""
    obj_name = ""

    """OpenDaylight LBaaS Driver for the V2 API

    This code is the backend implementation for the OpenDaylight
    LBaaS V2 driver for OpenStack Neutron.
    """

    @log_helpers.log_method_call
    def __init__(self, driver, client, obj_type):
        super(OpenDaylightManager, self).__init__(driver)
        self.client = client
        self.obj_type = obj_type
        self.url_path = LBAAS + '/' + obj_type
        self.obj_name = obj_type[:-1]

    @log_helpers.log_method_call
    @driver_base.driver_op
    def create(self, context, obj):
        self.client.sendjson('post', self.url_path,
                             {self.obj_name: obj.to_api_dict()})

    @log_helpers.log_method_call
    @driver_base.driver_op
    def update(self, context, obj):
        self.client.sendjson('put', self.url_path + '/' + obj.id,
                             {self.obj_name: obj.to_api_dict()})

    @log_helpers.log_method_call
    @driver_base.driver_op
    def delete(self, context, obj):
        self.client.sendjson('delete', self.url_path + '/' + obj.id, None)


class ODLLoadBalancerManager(OpenDaylightManager,
                             driver_base.BaseLoadBalancerManager):

    @log_helpers.log_method_call
    def __init__(self, driver, client):
        super(ODLLoadBalancerManager, self).__init__(
            driver, client, odl_const.ODL_LOADBALANCERS)

    @log_helpers.log_method_call
    @driver_base.driver_op
    def refresh(self, context, lb):
        # TODO(lijingjing): implement this method
        # This is intended to trigger the backend to check and repair
        # the state of this load balancer and all of its dependent objects
        pass

    @log_helpers.log_method_call
    @driver_base.driver_op
    def stats(self, context, lb):
        # TODO(lijingjing): implement this method
        pass

    # NOTE(yamahata): workaround for pylint
    # pylint raise false positive of abstract-class-instantiated.
    # method resolution order is as follows and db_delete_method is resolved
    # by BaseLoadBalancerManager. However pylint complains as this
    # class is still abstract class
    # mro:
    # ODLLoadBalancerManager
    # OpenDaylightManager
    # neutron_lbaas.drivers.driver_base.BaseLoadBalancerManager
    # neutron_lbaas.drivers.driver_mixins.BaseRefreshMixin
    # neutron_lbaas.drivers.driver_mixins.BaseStatsMixin
    # neutron_lbaas.drivers.driver_mixins.BaseManagerMixin
    # __builtin__.object
    @property
    def db_delete_method(self):
        return driver_base.BaseLoadBalancerManager.db_delete_method


class ODLListenerManager(OpenDaylightManager,
                         driver_base.BaseListenerManager):

    @log_helpers.log_method_call
    def __init__(self, driver, client):
        super(ODLListenerManager, self).__init__(
            driver, client, odl_const.ODL_LISTENERS)


class ODLPoolManager(OpenDaylightManager,
                     driver_base.BasePoolManager):

    @log_helpers.log_method_call
    def __init__(self, driver, client):
        super(ODLPoolManager, self).__init__(
            driver, client, odl_const.ODL_POOLS)


class ODLMemberManager(OpenDaylightManager,
                       driver_base.BaseMemberManager):

    # NOTE:It is for lbaas v2 api but using v1 mechanism of networking-odl.

    @log_helpers.log_method_call
    def __init__(self, driver, client):
        super(ODLMemberManager, self).__init__(
            driver, client, odl_const.ODL_MEMBERS)

    @log_helpers.log_method_call
    @driver_base.driver_op
    def create(self, context, obj):
        self.client.sendjson('post', self._member_url(obj),
                             {self.obj_name: obj.to_api_dict()})

    @log_helpers.log_method_call
    @driver_base.driver_op
    def update(self, context, obj):
        self.client.sendjson('put', self._member_url(obj) + '/' + obj.id,
                             {self.obj_name: obj.to_api_dict()})

    @log_helpers.log_method_call
    @driver_base.driver_op
    def delete(self, context, obj):
        self.client.sendjson('delete',
                             self._member_url(obj) + '/' + obj.id, None)

    def _member_url(self, obj):
        return (LBAAS + '/' + odl_const.ODL_POOLS + '/' + obj.pool_id + '/' +
                odl_const.ODL_MEMBERS)


class ODLHealthMonitorManager(OpenDaylightManager,
                              driver_base.BaseHealthMonitorManager):

    @log_helpers.log_method_call
    def __init__(self, driver, client):
        super(ODLHealthMonitorManager, self).__init__(
            driver, client, odl_const.ODL_HEALTHMONITORS)



# Copyright Intel 2016
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures

from neutron.agent.common import utils
from neutron.tests.common import net_helpers


class FakeMachine(fixtures.Fixture):
    """Create a fake machine.

    :ivar bridge: bridge on which the fake machine is bound
    :ivar ip_cidr: fake machine ip_cidr
    :type ip_cidr: str

    :ivar namespace: namespace emulating the machine
    :type namespace: str
    :ivar port: port binding the namespace to the bridge
    :type port: dict

    """

    def __init__(self, bridge, port):
        self.bridge = bridge
        self.port = port

    def _setUp(self):
        ns_fixture = self.useFixture(
            net_helpers.NamespaceFixture())
        self.namespace = ns_fixture.name

    def execute(self, cmds, run_as_root=True):
        ns_params = ['ip', 'netns', 'exec', self.namespace]
        cmd = ns_params + list(cmds)
        kwargs = {'run_as_root': run_as_root}
        return utils.execute(cmd, **kwargs)

    def assert_arping(self, dst_ip):
        return self.execute(
            ['arping', '-c', 3, '-I', self.interface_id, dst_ip])

    def assert_ping(self, dst_ip):
        return self.execute(['ping', '-c', 3, '-W', 1, dst_ip])

    @property
    def port_id(self):
        return self.port['port']['id']

    @property
    def interface_id(self):
        return 'ifce' + self.port_id[
            :net_helpers.OVSPortFixture.NIC_NAME_LEN - 3]

    @property
    def mac_address(self):
        return self.port['port']['mac_address']

    @property
    def ip_cidr(self):
        return self.port['port']['fixed_ips'][0]['ip_address'] + '/24'

    def set_mac_address(self):
        utils.execute(['ip', 'link', 'set', self.interface_id,
                       'netns', self.namespace],
                      run_as_root=True)

        self.execute(['ip', 'link', 'set', 'dev',
                      self.interface_id, 'address', self.mac_address])

    def set_ip_address(self):
        self.execute(['ip', 'address', 'add', self.ip_cidr,
                      'dev', self.interface_id])

    def set_dev_up(self):
        self.execute(['ip', 'link', 'set', 'dev',
                      self.interface_id, 'up'])

    def _create_ovs_vif_port(self, instance_id):
        return utils.execute(['ovs-vsctl', 'add-port', self.bridge,
                              self.interface_id, '--', 'set',
                              'Interface', self.interface_id,
                              'external-ids:iface-id=%s' % self.port_id,
                              'external-ids:iface-status=active',
                              'external-ids:attached-mac=%s'
                              % self.mac_address,
                              'external-ids:vm-uuid=%s' % instance_id,
                              'type=internal'], run_as_root=True)

    def set_address(self):
        self.set_mac_address()
        self.set_ip_address()
        self.set_dev_up()
